@@ -20,6 +20,22 @@ from sequifier.helpers import normalize_path
 def load_hyperparameter_search_config(
     config_path: str, on_unprocessed: bool
 ) -> "HyperparameterSearch":
+    """Load a hyperparameter search configuration from a YAML file.
+
+    This function reads a YAML configuration file, processes it to include
+    data-driven configurations if needed, and returns a HyperparameterSearch
+    object.
+
+    Args:
+        config_path: The path to the hyperparameter search configuration file.
+        on_unprocessed: A boolean flag indicating whether the configuration is
+            for unprocessed data. If False, it will load and integrate
+            data-driven configurations.
+
+    Returns:
+        An instance of the HyperparameterSearch class, populated with the
+        configuration from the file.
+    """
     with open(config_path, "r") as f:
         config_values = yaml.safe_load(f)
 
@@ -79,7 +95,26 @@ def load_hyperparameter_search_config(
 
 
 class TrainingSpecHyperparameterSampling(BaseModel):
-    """Pydantic model for training specifications."""
+    """Pydantic model for training specification hyperparameter sampling.
+
+    Attributes:
+        device: The device to train on (e.g., 'cuda', 'cpu').
+        epochs: A list of possible numbers of epochs to train for.
+        log_interval: The interval in batches for logging.
+        class_share_log_columns: Columns for which to log class share.
+        early_stopping_epochs: Number of epochs for early stopping.
+        iter_save: Interval in epochs for saving model checkpoints.
+        batch_size: A list of possible batch sizes.
+        lr: A list of possible learning rates.
+        criterion: A dictionary mapping target columns to loss functions.
+        class_weights: Optional dictionary mapping columns to class weights.
+        accumulation_steps: A list of possible gradient accumulation steps.
+        dropout: A list of possible dropout rates.
+        loss_weights: Optional dictionary mapping columns to loss weights.
+        optimizer: A list of possible optimizer configurations.
+        scheduler: A list of possible scheduler configurations.
+        continue_training: Flag to continue training from a checkpoint.
+    """
 
     device: str
     epochs: list[int]
@@ -105,6 +140,17 @@ class TrainingSpecHyperparameterSampling(BaseModel):
     continue_training: bool = True
 
     def __init__(self, **kwargs):
+        """Initialize the TrainingSpecHyperparameterSampling instance.
+
+        This method initializes the Pydantic BaseModel and then processes the
+        optimizer and scheduler configurations from the provided keyword
+        arguments, converting them into DotDict objects.
+
+        Args:
+            **kwargs: Keyword arguments that correspond to the attributes of this
+                class. The 'optimizer' and 'scheduler' arguments are expected
+                to be lists of dictionaries.
+        """
         super().__init__(
             **{k: v for k, v in kwargs.items() if k not in ["optimizer", "scheduler"]}
         )
@@ -128,6 +174,16 @@ class TrainingSpecHyperparameterSampling(BaseModel):
         return v
 
     def random_sample(self):
+        """Randomly sample a set of training hyperparameters.
+
+        This method selects a random combination of hyperparameters from the
+        defined lists of possibilities. It ensures that learning rates and
+        schedulers are paired correctly.
+
+        Returns:
+            A TrainingSpecModel instance populated with a randomly sampled set of
+            hyperparameters.
+        """
         lr_and_scheduler_index = np.random.randint(len(self.lr))
         optimizer_index = np.random.randint(len(self.optimizer))
         batch_size = np.random.choice(self.batch_size)
@@ -157,6 +213,18 @@ class TrainingSpecHyperparameterSampling(BaseModel):
         )
 
     def grid_sample(self, i):
+        """Select a set of training hyperparameters based on a grid search index.
+
+        This method generates a grid of all possible hyperparameter combinations
+        and selects the combination at the given index.
+
+        Args:
+            i: The index of the hyperparameter combination to select from the grid.
+
+        Returns:
+            A TrainingSpecModel instance populated with the selected set of
+            hyperparameters.
+        """
         hyperparameter_combinations = list(
             product(
                 np.arange(len(self.lr)),
@@ -193,6 +261,14 @@ class TrainingSpecHyperparameterSampling(BaseModel):
         )
 
     def n_combinations(self):
+        """Calculate the total number of hyperparameter combinations.
+
+        This method computes the total number of unique hyperparameter sets that
+        can be generated by the grid search.
+
+        Returns:
+            The total number of possible hyperparameter combinations.
+        """
         return (
             len(self.lr)
             * len(self.batch_size)
@@ -203,7 +279,15 @@ class TrainingSpecHyperparameterSampling(BaseModel):
 
 
 class ModelSpecHyperparameterSampling(BaseModel):
-    """Pydantic model for model specifications."""
+    """Pydantic model for model specification hyperparameter sampling.
+
+    Attributes:
+        d_model: A list of possible numbers of expected features in the input.
+        d_model_by_column: A list of possible embedding dimensions for each input column.
+        nhead: A list of possible numbers of heads in the multi-head attention models.
+        d_hid: A list of possible dimensions of the feedforward network model.
+        nlayers: A list of possible numbers of layers in the transformer model.
+    """
 
     d_model: list[int]
     d_model_by_column: Optional[list[dict[str, int]]]
@@ -224,6 +308,16 @@ class ModelSpecHyperparameterSampling(BaseModel):
         return v
 
     def random_sample(self):
+        """Randomly sample a set of model hyperparameters.
+
+        This method selects a random combination of model hyperparameters from the
+        defined lists of possibilities. It ensures that d_model, d_model_by_column,
+        and nhead are paired correctly.
+
+        Returns:
+            A ModelSpecModel instance populated with a randomly sampled set of
+            hyperparameters.
+        """
         d_model_index = np.random.randint(len(self.d_model))
         d_model_by_column = (
             None
@@ -244,6 +338,18 @@ class ModelSpecHyperparameterSampling(BaseModel):
         )
 
     def grid_sample(self, i):
+        """Select a set of model hyperparameters based on a grid search index.
+
+        This method generates a grid of all possible model hyperparameter
+        combinations and selects the combination at the given index.
+
+        Args:
+            i: The index of the hyperparameter combination to select from the grid.
+
+        Returns:
+            A ModelSpecModel instance populated with the selected set of
+            hyperparameters.
+        """
         hyperparameter_combinations = list(
             product(np.arange(len(self.d_model)), self.d_hid, self.nlayers)
         )
@@ -267,11 +373,46 @@ class ModelSpecHyperparameterSampling(BaseModel):
         )
 
     def n_combinations(self):
+        """Calculate the total number of model hyperparameter combinations.
+
+        This method computes the total number of unique model hyperparameter sets
+        that can be generated by the grid search.
+
+        Returns:
+            The total number of possible model hyperparameter combinations.
+        """
         return len(self.d_model) * len(self.d_hid) * len(self.nlayers)
 
 
 class HyperparameterSearch(BaseModel):
-    """Pydantic model for training configuration."""
+    """Pydantic model for hyperparameter search configuration.
+
+    Attributes:
+        project_path: The path to the sequifier project directory.
+        ddconfig_path: The path to the data-driven configuration file.
+        hp_search_name: The name for the hyperparameter search.
+        search_strategy: The search strategy, either "sample" or "grid".
+        n_samples: The number of samples to draw for the search.
+        model_config_write_path: The path to write the model configurations to.
+        training_data_path: The path to the training data.
+        validation_data_path: The path to the validation data.
+        read_format: The file format of the input data.
+        selected_columns: A list of lists of columns to be used for training.
+        column_types: A list of dictionaries mapping columns to their types.
+        categorical_columns: A list of lists of categorical columns.
+        real_columns: A list of lists of real-valued columns.
+        target_columns: The list of target columns for model training.
+        target_column_types: A dictionary mapping target columns to their types.
+        id_maps: A dictionary mapping categorical values to their indexed representation.
+        seq_length: A list of possible sequence lengths.
+        n_classes: The number of classes for each categorical column.
+        inference_batch_size: The batch size for inference.
+        export_onnx: If True, exports the model in ONNX format.
+        export_pt: If True, exports the model using torch.save.
+        export_with_dropout: If True, exports the model with dropout enabled.
+        model_hyperparameter_sampling: The sampling configuration for model hyperparameters.
+        training_hyperparameter_sampling: The sampling configuration for training hyperparameters.
+    """
 
     project_path: str
     ddconfig_path: str
@@ -311,6 +452,18 @@ class HyperparameterSearch(BaseModel):
         return v
 
     def random_sample(self, i):
+        """Randomly sample a full training configuration.
+
+        This method generates a complete training configuration by randomly
+        sampling model and training hyperparameters, as well as selecting a
+        column set and sequence length.
+
+        Args:
+            i: The index of the sample, used to create a unique model name.
+
+        Returns:
+            A TrainModel instance populated with a randomly sampled configuration.
+        """
         model_spec = self.model_hyperparameter_sampling.random_sample()
         training_spec = self.training_hyperparameter_sampling.random_sample()
         selected_columns_index = np.random.randint(len(self.selected_columns))
@@ -342,6 +495,17 @@ class HyperparameterSearch(BaseModel):
         )
 
     def grid_sample(self, i):
+        """Select a full training configuration based on a grid search index.
+
+        This method generates a grid of all possible configurations and selects
+        the configuration at the given index.
+
+        Args:
+            i: The index of the configuration to select from the grid.
+
+        Returns:
+            A TrainModel instance populated with the selected configuration.
+        """
         model_hyperparamter_sample = self.model_hyperparameter_sampling.n_combinations()
         training_hyperparamter_sample = (
             self.training_hyperparameter_sampling.n_combinations()
@@ -387,6 +551,20 @@ class HyperparameterSearch(BaseModel):
         )
 
     def sample(self, i):
+        """Sample a configuration based on the specified search strategy.
+
+        This method delegates to either random_sample or grid_sample based on
+        the `search_strategy` attribute.
+
+        Args:
+            i: The index of the sample or grid combination to generate.
+
+        Returns:
+            A TrainModel instance with a generated configuration.
+
+        Raises:
+            Exception: If the search_strategy is not 'sample' or 'grid'.
+        """
         if self.search_strategy == "sample":
             return self.random_sample(i)
         elif self.search_strategy == "grid":
@@ -395,6 +573,14 @@ class HyperparameterSearch(BaseModel):
             raise Exception(f"{self.search_strategy} invalid")
 
     def n_combinations(self):
+        """Calculate the total number of possible configurations.
+
+        This method computes the total number of unique configurations that can be
+        generated by a grid search over all defined hyperparameters.
+
+        Returns:
+            The total number of possible hyperparameter configurations.
+        """
         return (
             len(self.selected_columns)
             * len(self.seq_length)
