@@ -8,10 +8,20 @@ from beartype import beartype
 from torch import Tensor
 
 PANDAS_TO_TORCH_TYPES = {
-    "Int64": torch.int64,
     "Float64": torch.float32,
-    "int64": torch.int64,
     "float64": torch.float32,
+    "Float32": torch.float32,
+    "float32": torch.float32,
+    "Float16": torch.float16,
+    "float16": torch.float16,
+    "Int64": torch.int64,
+    "int64": torch.int64,
+    "Int32": torch.int32,
+    "int32": torch.int32,
+    "Int16": torch.int16,
+    "int16": torch.int16,
+    "Int8": torch.int8,
+    "int8": torch.int8,
 }
 
 
@@ -263,7 +273,7 @@ class LogFile:
     """
 
     @beartype
-    def __init__(self, path: str, open_mode: str, rank: Optional[int] = None):
+    def __init__(self, path: str, rank: Optional[int] = None):
         """Initializes the LogFile and opens log files.
 
         The `path` argument should be a template containing "[NUMBER]",
@@ -273,7 +283,6 @@ class LogFile:
         Args:
             path: The path template for the log files (e.g.,
                 "run_log_[NUMBER].txt").
-            open_mode: The mode for opening the log files (e.g., "a", "w").
             rank: The rank of the current process (e.g., in distributed
                 training). If None or 0, high-level messages will be
                 printed to stdout.
@@ -281,8 +290,7 @@ class LogFile:
         self.rank = rank
         self.levels = [2, 3]
         self._files = {
-            level: open(path.replace("[NUMBER]", str(level)), open_mode)
-            for level in self.levels
+            level: path.replace("[NUMBER]", str(level)) for level in self.levels
         }
         self._path = path
 
@@ -305,18 +313,12 @@ class LogFile:
         """
         for level2 in self.levels:
             if level2 <= level:
-                self._files[level2].write(f"{string}\n")
-                self._files[level2].flush()
+                with open(self._files[level2], "a+", encoding="utf-8") as f:
+                    f.write(f"{string}\n")
         if level >= 3:
             if self.rank is None or self.rank == 0:
                 print(string)
 
-    @beartype
-    def close(self) -> None:
-        """Closes all open log file handlers."""
-        for file in self._files.values():
-            file.close()
-
 
 @beartype
 def normalize_path(path: str, project_path: str) -> str:
