@@ -135,16 +135,11 @@ def infer_worker(
     """
     print(f"[INFO] Reading data from '{config.data_path}'...")
     # Step 1: Use Polars for data ingestion
+    dataset = None
     if config.read_format == "parquet":
         dataset = [pl.read_parquet(config.data_path)]
     elif config.read_format == "csv":
         dataset = [pl.read_csv(config.data_path)]
-    elif config.read_format == "pt":
-        assert percentage_limits is not None
-        start_pct, end_pct = percentage_limits
-        dataset = load_pt_dataset(config.data_path, start_pct, end_pct)
-    else:
-        raise Exception(f"{config.read_format = } not in ['parquet', 'csv', 'pt']")
 
     model_paths = (
         config.model_path
@@ -152,6 +147,14 @@ def infer_worker(
         else [config.model_path]
     )
     for model_path in model_paths:
+        if config.read_format == "pt":
+            assert percentage_limits is not None
+            start_pct, end_pct = percentage_limits
+            dataset = load_pt_dataset(config.data_path, start_pct, end_pct)
+
+        if dataset is None:
+            raise Exception(f"{config.read_format = } not in ['parquet', 'csv', 'pt']")
+
         inferer = Inferer(
             config.model_type,
             model_path,
@@ -203,8 +206,8 @@ def infer_embedding(
     of DataFrames or an iterator of tensors). For each data chunk, it
     calls the appropriate function (`get_embeddings` or `get_embeddings_pt`)
     to generate embeddings. It then formats these embeddings into a
-    Polars DataFrame, associating them with their `sequenceId`, and
-    writes the resulting DataFrame to the configured output path.
+    Polars DataFrame, associating them with their `sequenceId` and
+    `subsequenceId`, and writes the resulting DataFrame to the configured output path.
 
     Args:
         config: The `InfererModel` configuration object.
@@ -225,16 +228,19 @@ def infer_embedding(
             mask = pl.arange(0, data.height, eager=True) % n_input_cols == 0
             embeddings = get_embeddings(config, inferer, data, column_types)
             sequence_ids_for_preds = data.get_column("sequenceId").filter(mask)
+            subsequence_ids_for_preds = data.get("subsequenceId").filter(mask)
         elif config.read_format == "pt":
-            sequences_dict, _, sequence_ids_tensor = data
+            sequences_dict, _, sequence_ids_tensor, subsequence_ids_tensor, _ = data
             embeddings = get_embeddings_pt(config, inferer, sequences_dict)
             sequence_ids_for_preds = sequence_ids_tensor.numpy()
+            subsequence_ids_for_preds = subsequence_ids_tensor.numpy()
         else:
             raise Exception("impossible")
 
         embeddings = pl.DataFrame(
             {
                 "sequenceId": sequence_ids_for_preds,
+                "subsequenceId": subsequence_ids_for_preds,
                 **dict(
                     zip(
                         [str(v) for v in range(embeddings.shape[1])],
@@ -286,7 +292,8 @@ def infer_generative(
     This function manages the generative inference workflow:
     1. Iterates through the dataset (chunks).
     2. Handles data preparation, including expanding data for autoregression
-       if configured (`expand_data_by_autoregression`).
+       if configured (`expand_data_by_autoregression`). It also calculates
+       the corresponding `itemPosition` for each prediction.
     3. Calls the correct function to get probabilities and predictions
        based on data format and autoregression settings (e.g.,
        `get_probs_preds_autoregression`, `get_probs_preds_pt`).
@@ -295,7 +302,7 @@ def infer_generative(
        - Inverts normalization for real-valued target columns.
     5. Saves probabilities to disk (if `config.output_probabilities` is True).
     6. Saves the final predictions to disk, formatted as a Polars DataFrame
-       with `sequenceId` and target columns.
+       with `sequenceId`, `itemPosition`, and target columns.
 
     Args:
         config: The `InfererModel` configuration object.
@@ -312,13 +319,18 @@ def infer_generative(
         if config.read_format in ["parquet", "csv"]:
             if config.selected_columns is not None:
                 data = subset_to_selected_columns(data, config.selected_columns)
+            n_input_cols = data.get_column("inputCol").n_unique()
             if not config.autoregression:
                 # For the non-autoregressive case, the old logic is still needed here
-                n_input_cols = data.get_column("inputCol").n_unique()
-                mask = pl.arange(0, data.height, eager=True) % n_input_cols == 0
                 # Apply the mask to the sequenceId column
+                mask = pl.arange(0, data.height, eager=True) % n_input_cols == 0
                 sequence_ids_for_preds = data.get_column("sequenceId").filter(mask)
+                item_positions_for_preds = (
+                    data.get_column("startItemPosition").filter(mask).to_numpy()
+                    + config.seq_length
+                )
                 probs, preds = get_probs_preds(config, inferer, data, column_types)
+
             else:
                 if config.autoregression_extra_steps is not None:
                     data = expand_data_by_autoregression(
@@ -326,13 +338,17 @@ def infer_generative(
                         config.autoregression_extra_steps,
                         config.seq_length,
                     )
-
+                mask = pl.arange(0, data.height, eager=True) % n_input_cols == 0
+                item_positions_for_preds = (
+                    data.get_column("startItemPosition").filter(mask).to_numpy()
+                    + config.seq_length
+                )
                 # Unpack the new third return value
                 probs, preds, sequence_ids_for_preds = get_probs_preds_autoregression(
                     config, inferer, data, column_types, config.seq_length
                 )
         elif config.read_format == "pt":
-            sequences_dict, _, sequence_ids_tensor = data
+            sequences_dict, _, sequence_ids_tensor, _, start_positions_tensor = data
             extra_steps = (
                 0
                 if config.autoregression_extra_steps is None
@@ -344,6 +360,14 @@ def infer_generative(
             sequence_ids_for_preds = np.repeat(
                 sequence_ids_tensor.numpy(), extra_steps + 1
             )
+            item_position_boundaries = zip(
+                list(start_positions_tensor + config.seq_length),
+                list(start_positions_tensor + config.seq_length + extra_steps + 1),
+            )
+            item_positions_for_preds = np.concatenate(
+                [np.arange(start, end) for start, end in item_position_boundaries],
+                axis=0,
+            )
         else:
             raise Exception("impossible")
 
@@ -409,6 +433,7 @@ def infer_generative(
         predictions = pl.DataFrame(
             {
                 "sequenceId": sequence_ids_for_preds,
+                "itemPosition": item_positions_for_preds,
                 **{
                     target_column: preds[target_column].flatten()
                     for target_column in inferer.target_columns
@@ -485,6 +510,8 @@ def expand_data_by_autoregression(
     for offset in range(1, autoregression_extra_steps + 1):
         future_df_lazy = last_obs_lazy.with_columns(
             (pl.col("subsequenceId") + offset).alias("subsequenceId")
+        ).with_columns(
+            (pl.col("startItemPosition") + offset).alias("startItemPosition")
         )
 
         # Correctly shift columns to make space for future predictions
@@ -525,14 +552,16 @@ def get_embeddings_pt(
 
     This function serves as a wrapper for `Inferer.infer_embedding` when
     the input data is already in PyTorch tensor format (from loading `.pt`
-    files). It converts the tensor dictionary to a NumPy array dictionary
-    before passing it to the inferer.
+    files which contain sequences, targets, sequence_ids, subsequence_ids,
+    and start_positions). It converts the tensor dictionary to a NumPy array
+    dictionary before passing it to the inferer.
 
     Args:
         config: The `InfererModel` configuration object (unused, but
             kept for consistent function signature).
         inferer: The initialized `Inferer` instance.
-        data: A dictionary mapping column/feature names to `torch.Tensor`s.
+        data: A dictionary mapping column/feature names to `torch.Tensor`s
+              (the sequences part loaded from the .pt file).
 
     Returns:
         A NumPy array containing the computed embeddings for the batch.
@@ -552,20 +581,23 @@ def get_probs_preds_pt(
     """Generates predictions from PyTorch tensor data, supporting autoregression.
 
     This function performs generative inference on a batch of PyTorch tensor
-    data. It implements an autoregressive loop:
-    1. Runs inference on the initial data `X`.
+    data loaded from `.pt` files (which contain sequences, targets,
+    sequence_ids, subsequence_ids, and start_positions). It implements an
+    autoregressive loop:
+    1. Runs inference on the initial data `X` (sequences).
     2. For each subsequent step (`i` in `extra_steps`):
        a. Creates the next input `X_next` by shifting the previous input
           `X` and appending the prediction from the last step.
        b. Runs inference on `X_next`.
     3. Collects and reshapes all predictions and probabilities from all
-       steps into a single flat batch.
+       steps into a single flat batch, ordered by original sample index, then by step.
 
     Args:
         config: The `InfererModel` configuration object, used to check
             `output_probabilities` and `selected_columns`.
         inferer: The initialized `Inferer` instance.
-        data: A dictionary mapping column/feature names to `torch.Tensor`s.
+        data: A dictionary mapping column/feature names to `torch.Tensor`s
+              (the sequences part loaded from the .pt file).
         extra_steps: The number of additional autoregressive steps to
             perform. A value of 0 means simple, non-autoregressive
             inference.
@@ -573,10 +605,10 @@ def get_probs_preds_pt(
     Returns:
         A tuple `(probs, preds)`:
             - `probs`: A dictionary mapping target columns to NumPy arrays
-              of probabilities, or `None` if `config.output_probabilities`
-              is False.
+              of probabilities, ordered by sample index then step,
+              or `None` if `config.output_probabilities` is False.
             - `preds`: A dictionary mapping target columns to NumPy arrays
-              of final predictions.
+              of final predictions, ordered by sample index then step.
     """
     target_cols = inferer.target_columns
 
@@ -1178,7 +1210,7 @@ class Inferer:
             A 2D NumPy array of the resulting embeddings.
         """
         assert x is not None
-        size = x[self.target_columns[0]].shape[0]
+        size = x[list(x.keys())[0]].shape[0]
         embedding = self.adjust_and_infer_embedding(x, size)
 
         return embedding
@@ -1224,7 +1256,7 @@ class Inferer:
             x is not None and len(set(x.keys()).difference(set(probs.keys()))) > 0
         ):  # type: ignore
             assert x is not None
-            size = x[self.target_columns[0]].shape[0]
+            size = x[list(x.keys())[0]].shape[0]
             if (
                 probs is not None
                 and len(set(x.keys()).difference(set(probs.keys()))) > 0
@@ -1370,7 +1402,7 @@ class Inferer:
             A list of dictionaries, where each dictionary is a single batch
             ready for inference.
         """
-        size = x[self.target_columns[0]].shape[0]
+        size = x[list(x.keys())[0]].shape[0]
         if size == self.inference_batch_size:
             return [x]
         elif size < self.inference_batch_size:
