@@ -75,6 +75,7 @@ class Preprocessor:
         n_cores: Optional[int],
         batches_per_file: int,
         process_by_file: bool,
+        subsequence_start_mode: str,
     ):
         """Initializes the Preprocessor with the given parameters.
 
@@ -135,7 +136,7 @@ class Preprocessor:
                 id_maps, n_classes, col_types, selected_columns_statistics
             )
 
-            schema = self._create_schema(col_types, seq_length)
+            schema = self._create_schema(col_types, seq_length + 1)
 
             data = data.sort(["sequenceId", "itemPosition"])
             n_batches = _process_batches_single_file(
@@ -153,6 +154,7 @@ class Preprocessor:
                 self.split_paths,
                 self.target_dir,
                 self.batches_per_file,
+                subsequence_start_mode,
             )
 
             if self.combine_into_single_file:
@@ -176,17 +178,20 @@ class Preprocessor:
                 delete_files(input_files)
 
         else:
-            n_classes, id_maps, selected_columns_statistics, col_types, data_columns = (
-                self._get_column_metadata_across_files(
-                    data_path, read_format, max_rows, selected_columns
-                )
+            (
+                files_to_process,
+                n_classes,
+                id_maps,
+                selected_columns_statistics,
+                col_types,
+                data_columns,
+            ) = self._get_column_metadata_across_files(
+                data_path, read_format, max_rows, selected_columns
             )
             self._export_metadata(
                 id_maps, n_classes, col_types, selected_columns_statistics
             )
-            schema = self._create_schema(col_types, seq_length)
-
-            files_to_process = self._get_files_to_process(data_path, read_format)
+            schema = self._create_schema(col_types, seq_length + 1)
 
             self._process_batches_multiple_files(
                 files_to_process,
@@ -205,32 +210,11 @@ class Preprocessor:
                 group_proportions,
                 write_format,
                 process_by_file,
+                subsequence_start_mode,
             )
 
         self._cleanup(write_format)
 
-    @beartype
-    def _get_files_to_process(self, data_path: str, read_format: str) -> list[str]:
-        """Finds all data files within a directory that match the read format.
-
-        This method recursively walks through the given `data_path` directory
-        and collects the full paths of all files that end with the specified
-        `read_format` (e.g., ".csv", ".parquet").
-
-        Args:
-            data_path: The path to the root data directory to search.
-            read_format: The file extension (e.g., "csv", "parquet") to look for.
-
-        Returns:
-            A list of string paths to the files that match the criteria.
-        """
-        paths_to_process = []
-        for root, dirs, files in os.walk(data_path):
-            for file in files:
-                if file.endswith(read_format):
-                    paths_to_process.append(os.path.join(root, file))
-        return paths_to_process
-
     @beartype
     def _create_schema(
         self, col_types: dict[str, str], seq_length: int
@@ -238,8 +222,9 @@ class Preprocessor:
         """Creates the Polars schema for the intermediate sequence DataFrame.
 
         This schema defines the structure of the DataFrame after sequence
-        extraction, which includes sequence identifiers, the input column name,
-        and columns for each item in the sequence (named '0', '1', ..., 'seq_length-1').
+        extraction, which includes sequence identifiers, the start item position
+        within the original sequence, the input column name, and columns for
+        each item in the sequence (named '0', '1', ..., 'seq_length-1').
 
         Args:
             col_types: A dictionary mapping data column names to their Polars
@@ -248,12 +233,13 @@ class Preprocessor:
 
         Returns:
             A dictionary defining the Polars schema. Keys are column names
-            (e.g., "sequenceId", "subsequenceId", "inputCol", "0", "1", ...)
+            (e.g., "sequenceId", "subsequenceId", "startItemPosition", "inputCol", "0", "1", ...)
             and values are Polars data types (e.g., `pl.Int64`).
         """
         schema = {
             "sequenceId": pl.Int64,
             "subsequenceId": pl.Int64,
+            "startItemPosition": pl.Int64,
             "inputCol": pl.String,
         }
 
@@ -282,6 +268,7 @@ class Preprocessor:
         max_rows: Optional[int],
         selected_columns: Optional[list[str]],
     ) -> tuple[
+        list[str],
         dict[str, int],
         dict[str, dict[Union[str, int], int]],
         dict[str, dict[str, float]],
@@ -324,15 +311,16 @@ class Preprocessor:
         n_rows_running_count = 0
         id_maps, selected_columns_statistics = {}, {}
         col_types, data_columns = None, None
-
+        files_to_process = []
         print(f"[INFO] Data path: {data_path}")
         for root, dirs, files in os.walk(data_path):
             print(f"[INFO] N Files : {len(files)}")
-            for file in files:
+            for file in sorted(list(files)):
                 if file.endswith(read_format) and (
                     max_rows is None or n_rows_running_count < max_rows
                 ):
                     print(f"[INFO] Preprocessing: reading {file}")
+                    files_to_process.append(os.path.join(root, file))
                     max_rows_inner = (
                         None
                         if max_rows is None
@@ -369,7 +357,9 @@ class Preprocessor:
         assert data_columns is not None
         n_classes = {col: len(id_maps[col]) + 1 for col in id_maps}
         assert col_types is not None
+        files_to_process = sorted(files_to_process)
         return (
+            files_to_process,
             n_classes,
             id_maps,
             selected_columns_statistics,
@@ -434,6 +424,7 @@ class Preprocessor:
         group_proportions: list[float],
         write_format: str,
         process_by_file: bool = True,
+        subsequence_start_mode: str = "distribute",
     ) -> None:
         """Processes batches of data from multiple files.
 
@@ -479,6 +470,7 @@ class Preprocessor:
                 target_dir=self.target_dir,
                 batches_per_file=self.batches_per_file,
                 combine_into_single_file=self.combine_into_single_file,
+                subsequence_start_mode=subsequence_start_mode,
             )
             input_files = create_file_paths_for_multiple_files2(
                 self.project_path,
@@ -521,6 +513,7 @@ class Preprocessor:
                 "target_dir": self.target_dir,
                 "batches_per_file": self.batches_per_file,
                 "combine_into_single_file": self.combine_into_single_file,
+                "subsequence_start_mode": subsequence_start_mode,
             }
 
             job_params = [
@@ -674,7 +667,7 @@ class Preprocessor:
         for file_path in pt_files:
             try:
                 # Load the tensor file to inspect its contents
-                sequences_dict, _, _ = torch.load(file_path)
+                sequences_dict, _, _, _, _ = torch.load(file_path)
                 if sequences_dict:
                     # All tensors in the dict have the same number of samples (batch size)
                     n_samples = sequences_dict[list(sequences_dict.keys())[0]].shape[0]
@@ -897,6 +890,7 @@ def _process_batches_multiple_files_inner(
     target_dir: str,
     batches_per_file: int,
     combine_into_single_file: bool,
+    subsequence_start_mode: str,
 ):
     """Inner function for processing batches of data from multiple files.
 
@@ -924,10 +918,14 @@ def _process_batches_multiple_files_inner(
         batches_per_file: The number of batches to process per file.
         combine_into_single_file: Whether to combine the output into a single file.
     """
+    n_files = len(file_paths)
+    assert n_files > 0
+    pad_width = len(str(n_files - 1))
     n_rows_running_count = 0
     for file_index, path in enumerate(file_paths):
         max_rows_inner = None if max_rows is None else max_rows - n_rows_running_count
         if max_rows_inner is None or max_rows_inner > 0:
+            file_index_str = str(file_index).zfill(pad_width)
             data = _load_and_preprocess_data(
                 path, read_format, selected_columns, max_rows_inner
             )
@@ -942,12 +940,12 @@ def _process_batches_multiple_files_inner(
 
             adjusted_split_paths = [
                 path.replace(
-                    data_name_root, f"{data_name_root}-{process_id}-{file_index}"
+                    data_name_root, f"{data_name_root}-{process_id}-{file_index_str}"
                 )
                 for path in split_paths
             ]
 
-            data_name_root_inner = f"{data_name_root}-{process_id}-{file_index}"
+            data_name_root_inner = f"{data_name_root}-{process_id}-{file_index_str}"
 
             n_batches = _process_batches_single_file(
                 project_path,
@@ -964,6 +962,7 @@ def _process_batches_multiple_files_inner(
                 adjusted_split_paths,
                 target_dir,
                 batches_per_file,
+                subsequence_start_mode,
             )
 
             if combine_into_single_file:
@@ -973,7 +972,7 @@ def _process_batches_multiple_files_inner(
                     len(group_proportions),
                     n_batches,
                     process_id,
-                    file_index,
+                    file_index_str,
                     data_name_root,
                     write_format,
                 )
@@ -985,7 +984,7 @@ def _process_batches_multiple_files_inner(
                     data_name_root,
                     write_format,
                     in_target_dir=True,
-                    pre_split_str=f"{process_id}-{file_index}",
+                    pre_split_str=f"{process_id}-{file_index_str}",
                 )
 
                 delete_files(input_files)
@@ -1009,6 +1008,7 @@ def _process_batches_single_file(
     split_paths: list[str],
     target_dir: str,
     batches_per_file: int,
+    subsequence_start_mode: str,
 ) -> int:
     """Processes batches of data from a single file.
 
@@ -1049,6 +1049,7 @@ def _process_batches_single_file(
             target_dir,
             write_format,
             batches_per_file,
+            subsequence_start_mode,
         )
         for process_id, (start, end) in enumerate(batch_limits)
         if (end - start) > 0
@@ -1224,13 +1225,14 @@ def process_and_write_data_pt(
     This function takes the long-format sequence DataFrame (`data`),
     aggregates it by `sequenceId` and `subsequenceId`, and pivots it
     so that each `inputCol` becomes its own column containing a list
-    of sequence items.
+    of sequence items. It also extracts the `startItemPosition`.
 
     It then converts these lists into NumPy arrays, splits them into
     `sequences` (all but last item) and `targets` (all but first item),
-    and converts them to PyTorch tensors. The final data
-    `(sequences_dict, targets_dict, sequence_ids_tensor)` is saved
-    to a .pt file using `torch.save`.
+    and converts them to PyTorch tensors along with sequence/subsequence IDs
+    and start positions. The final data tuple
+    `(sequences_dict, targets_dict, sequence_ids_tensor, subsequence_ids_tensor, start_item_positions_tensor)`
+    is saved to a .pt file using `torch.save`.
 
     Args:
         data: The long-format Polars DataFrame of extracted sequences.
@@ -1243,7 +1245,7 @@ def process_and_write_data_pt(
     if data.is_empty():
         return
 
-    sequence_cols = [str(c) for c in range(seq_length - 1, -1, -1)]
+    sequence_cols = [str(c) for c in range(seq_length, -1, -1)]
 
     all_feature_cols = data.get_column("inputCol").unique().to_list()
 
@@ -1253,7 +1255,7 @@ def process_and_write_data_pt(
         .flatten()
         .alias(f"seq_{col_name}")
         for col_name in all_feature_cols
-    ]
+    ] + [pl.col("startItemPosition").first().alias("startItemPosition")]
 
     aggregated_data = (
         data.group_by(["sequenceId", "subsequenceId"])
@@ -1267,7 +1269,12 @@ def process_and_write_data_pt(
     sequence_ids_tensor = torch.tensor(
         aggregated_data.get_column("sequenceId").to_numpy(), dtype=torch.int64
     )
-
+    subsequence_ids_tensor = torch.tensor(
+        aggregated_data.get_column("subsequenceId").to_numpy(), dtype=torch.int64
+    )
+    start_item_positions_tensor = torch.tensor(
+        aggregated_data.get_column("startItemPosition").to_numpy(), dtype=torch.int64
+    )
     sequences_dict = {}
     targets_dict = {}
 
@@ -1285,7 +1292,13 @@ def process_and_write_data_pt(
         return
 
     print(f"[INFO] Writing preprocessed data to '{path}'...")
-    data_to_save = (sequences_dict, targets_dict, sequence_ids_tensor)
+    data_to_save = (
+        sequences_dict,
+        targets_dict,
+        sequence_ids_tensor,
+        subsequence_ids_tensor,
+        start_item_positions_tensor,
+    )
     torch.save(data_to_save, path)
 
 
@@ -1295,7 +1308,7 @@ def _write_accumulated_sequences(
     split_path: str,
     write_format: str,
     process_id: int,
-    file_index: int,
+    file_index_str: str,
     target_dir: str,
     seq_length: int,
     col_types: dict[str, str],
@@ -1324,7 +1337,7 @@ def _write_accumulated_sequences(
 
     # Construct a unique filename for the batched file
     split_path_batch_seq = split_path.replace(
-        f".{write_format}", f"-{process_id}-{file_index}.{write_format}"
+        f".{write_format}", f"-{process_id}-{file_index_str}.{write_format}"
     )
     out_path = insert_top_folder(split_path_batch_seq, target_dir)
 
@@ -1348,6 +1361,7 @@ def preprocess_batch(
     target_dir: str,
     write_format: str,
     batches_per_file: int,
+    subsequence_start_mode: str,
 ) -> None:
     """Processes a batch of data.
 
@@ -1374,6 +1388,7 @@ def preprocess_batch(
         sequences_by_split = {i: [] for i in range(len(split_paths))}
         file_indices = {i: 0 for i in range(len(split_paths))}
 
+        pad_width = len(str(math.ceil(len(sequence_ids) / batches_per_file) + 1))
         for i, sequence_id in enumerate(sequence_ids):
             data_subset = batch.filter(pl.col("sequenceId") == sequence_id)
             group_bounds = get_group_bounds(data_subset, group_proportions)
@@ -1385,6 +1400,7 @@ def preprocess_batch(
                         seq_length,
                         seq_step_sizes[i],
                         data_columns,
+                        subsequence_start_mode,
                     )
                 )
                 for i, (lb, ub) in enumerate(group_bounds)
@@ -1401,7 +1417,7 @@ def preprocess_batch(
                         split_paths[group],
                         write_format,
                         process_id,
-                        file_indices[group],
+                        str(file_indices[group]).zfill(pad_width),
                         target_dir,
                         seq_length,
                         col_types,
@@ -1417,7 +1433,7 @@ def preprocess_batch(
                 split_paths[group],
                 write_format,
                 process_id,
-                file_indices[group],
+                str(file_indices[group]).zfill(pad_width),
                 target_dir,
                 seq_length,
                 col_types,
@@ -1436,6 +1452,7 @@ def preprocess_batch(
                         seq_length,
                         seq_step_sizes[j],
                         data_columns,
+                        subsequence_start_mode,
                     )
                 )
                 for j, (lb, ub) in enumerate(group_bounds)
@@ -1476,14 +1493,16 @@ def extract_sequences(
     seq_length: int,
     seq_step_size: int,
     columns: list[str],
+    subsequence_start_mode: str,
 ) -> pl.DataFrame:
     """Extracts subsequences from a DataFrame of full sequences.
 
     This function takes a DataFrame where each row contains all items
     for a single `sequenceId`. It iterates through each `sequenceId`,
     extracts all possible subsequences of `seq_length` using the
-    specified `seq_step_size`, and formats them into a new, long-format
-    DataFrame that conforms to the provided `schema`.
+    specified `seq_step_size`, calculates the starting position of each
+    subsequence within the original sequence, and formats them into a new,
+    long-format DataFrame that conforms to the provided `schema`.
 
     Args:
         data: The input Polars DataFrame, grouped by "sequenceId".
@@ -1495,11 +1514,15 @@ def extract_sequences(
 
     Returns:
         A new, long-format Polars DataFrame containing the extracted
-        subsequences, matching the provided `schema`.
+        subsequences, matching the provided `schema`. Includes columns
+        for `sequenceId`, `subsequenceId`, `startItemPosition`, `inputCol`,
+        and the sequence items ('0', '1', ...).
     """
     if data.is_empty():
         return pl.DataFrame(schema=schema)
 
+    print(f"{data.shape = }")
+
     raw_sequences = data.group_by("sequenceId", maintain_order=True).agg(
         [pl.col(c) for c in columns]
     )
@@ -1509,16 +1532,26 @@ def extract_sequences(
         in_seq_lists_only = {col: in_row[col] for col in columns}
 
         subsequences = extract_subsequences(
-            in_seq_lists_only, seq_length, seq_step_size, columns
+            in_seq_lists_only,
+            seq_length,
+            seq_step_size,
+            columns,
+            subsequence_start_mode,
         )
 
         for subsequence_id in range(len(subsequences[columns[0]])):
             for col, subseqs in subsequences.items():
-                row = [in_row["sequenceId"], subsequence_id, col] + subseqs[
-                    subsequence_id
-                ]
-                assert len(row) == (seq_length + 3), f"{row = }"
+                row = [
+                    in_row["sequenceId"],
+                    subsequence_id,
+                    subsequence_id * seq_step_size,
+                    col,
+                ] + subseqs[subsequence_id]
+                assert len(row) == (seq_length + 5), f"{row = }"
                 rows.append(row)
+    print(f"{len(rows) = }")
+    print(f"{[len(r) for r in rows] = }")
+    print(schema)
 
     sequences = pl.DataFrame(
         rows,
@@ -1530,7 +1563,7 @@ def extract_sequences(
 
 @beartype
 def get_subsequence_starts(
-    in_seq_length: int, seq_length: int, seq_step_size: int
+    in_seq_length: int, seq_length: int, seq_step_size: int, subsequence_start_mode: str
 ) -> np.ndarray:
     """Calculates the start indices for extracting subsequences.
 
@@ -1548,15 +1581,38 @@ def get_subsequence_starts(
     Returns:
         A numpy array of integer start indices for each subsequence.
     """
-    nseq_adjusted = math.ceil((in_seq_length - seq_length) / seq_step_size)
-    seq_step_size_adjusted = math.floor(
-        (in_seq_length - seq_length) / max(1, nseq_adjusted)
-    )
-    increments = [0] + [max(1, seq_step_size_adjusted)] * nseq_adjusted
-    while np.sum(increments) < (in_seq_length - seq_length):
-        increments[np.argmin(increments[1:]) + 1] += 1
+    assert subsequence_start_mode in [
+        "distribute",
+        "exact",
+    ], f"{subsequence_start_mode = } not in ['distribute', 'exact']"
+    if subsequence_start_mode == "distribute":
+        last_available_start = in_seq_length - (seq_length + 1)
+        starts = np.arange(0, last_available_start + seq_step_size, seq_step_size)
+        starts[-1] = last_available_start
+        print(f"{starts = }")
 
-    return np.cumsum(increments)
+        if len(starts) > 2:
+            while True:
+                starts_delta = starts[1:] - starts[:-1]
+                if np.max(starts_delta) - np.min(starts_delta) > 1:
+                    starts[np.argmin(starts_delta) + 1] -= 1
+                else:
+                    print(f"{starts = }")
+                    return starts
+        print(f"{starts = }")
+        return starts
+
+    if subsequence_start_mode == "exact":
+        assert (
+            ((in_seq_length - 1) - seq_length) % seq_step_size == 0
+        ), f"'exact' can only be used if: ((in_seq_length - 1) - seq_length) % seq_step_size == 0, {(in_seq_length -1) = }, {seq_length = }, {seq_step_size = }"
+        last_possible_start = (
+            in_seq_length - (seq_length - 1) - 1
+        )  # the latter '-1' is to translate to index
+        return np.arange(
+            0, last_possible_start + 1, seq_step_size
+        )  # the '+1' is to make it inclusive
+    return np.array([])
 
 
 @beartype
@@ -1565,6 +1621,7 @@ def extract_subsequences(
     seq_length: int,
     seq_step_size: int,
     columns: list[str],
+    subsequence_start_mode: str,
 ) -> dict[str, list[list[Union[float, int]]]]:
     """Extracts subsequences from a dictionary of sequence lists.
 
@@ -1590,17 +1647,22 @@ def extract_subsequences(
         return {col: [] for col in columns}
 
     in_seq_len = len(in_seq[columns[0]])
-    if in_seq_len < seq_length:
-        pad_len = seq_length - in_seq_len
+    if in_seq_len < (seq_length + 1):
+        pad_len = (seq_length + 1) - in_seq_len
         in_seq = {col: ([0] * pad_len) + in_seq[col] for col in columns}
     in_seq_length = len(in_seq[columns[0]])
 
     subsequence_starts = get_subsequence_starts(
-        in_seq_length, seq_length, seq_step_size
+        in_seq_length, seq_length, seq_step_size, subsequence_start_mode
     )
+    subsequence_starts_diff = subsequence_starts[1:] - subsequence_starts[:-1]
+    assert np.all(
+        subsequence_starts_diff <= seq_step_size
+    ), f"Diff of {subsequence_starts = }, {subsequence_starts_diff = } larger than {seq_step_size = }"
+    print(f"{subsequence_starts = }")
 
     return {
-        col: [list(in_seq[col][i : i + seq_length]) for i in subsequence_starts]
+        col: [list(in_seq[col][i : i + seq_length + 1]) for i in subsequence_starts]
         for col in columns
     }
 
@@ -1663,7 +1725,7 @@ def create_file_paths_for_multiple_files1(
     n_splits: int,
     n_batches: int,
     process_id: int,
-    file_index: int,
+    file_index_str: str,
     dataset_name: str,
     write_format: str,
 ) -> dict[int, list[str]]:
@@ -1674,7 +1736,7 @@ def create_file_paths_for_multiple_files1(
     *before* they are combined.
 
     The naming pattern is:
-    `{dataset_name}-{process_id}-{file_index}-split{split}-{batch_id}.{write_format}`
+    `{dataset_name}-{process_id}-{file_index_str}-split{split}-{batch_id}.{write_format}`
 
     Args:
         project_path: The path to the sequifier project directory.
@@ -1682,7 +1744,7 @@ def create_file_paths_for_multiple_files1(
         n_splits: The number of data splits.
         n_batches: The number of batches created by the process.
         process_id: The ID of the multiprocessing worker.
-        file_index: The index of the file being processed by this worker.
+        file_index_str: The index of the file being processed by this worker.
         dataset_name: The root name of the dataset.
         write_format: The file extension.
 
@@ -1697,7 +1759,7 @@ def create_file_paths_for_multiple_files1(
                 project_path,
                 "data",
                 target_dir,
-                f"{dataset_name}-{process_id}-{file_index}-split{split}-{batch_id}.{write_format}",
+                f"{dataset_name}-{process_id}-{file_index_str}-split{split}-{batch_id}.{write_format}",
             )
             for batch_id in range(n_batches)
         ]
@@ -1784,13 +1846,15 @@ def create_file_paths_for_multiple_files2(
         intermediate combined file paths (str) for that split.
     """
     files = {}
+    n_files_max = max(n_files.values()) if n_files else 1
+    pad_width = len(str(n_files_max - 1))
     for split in range(n_splits):
         files_for_split = [
             os.path.join(
                 project_path,
                 "data",
                 target_dir,
-                f"{dataset_name}-{process_id}-{file_index}-split{split}.{write_format}",
+                f"{dataset_name}-{process_id}-{str(file_index).zfill(pad_width)}-split{split}.{write_format}",
             )
             for process_id in range(n_processes)
             for file_index in range(n_files[process_id])
