@@ -162,7 +162,8 @@ def train_worker(rank: int, world_size: int, config: TrainModel, from_folder: bo
     if config.training_spec.distributed:
         model = DDP(model, device_ids=[rank], find_unused_parameters=True)
 
-    model = torch.compile(model)
+    if config.training_spec.device.startswith("cuda"):
+        model = torch.compile(model)
 
     # 3. Start training
     # When using DDP, the original model is accessed via the .module attribute
@@ -215,7 +216,7 @@ def format_number(number: Union[int, float, np.float32]) -> str:
     elif number == 0:
         order_of_magnitude = 0
     else:
-        order_of_magnitude = math.floor(math.log(number, 10))
+        order_of_magnitude = math.floor(math.log(np.abs(number), 10))
 
     number_adjusted = number * (10 ** (-order_of_magnitude))
     return f"{number_adjusted:5.2f}e{order_of_magnitude}"
@@ -287,6 +288,8 @@ class TransformerModel(nn.Module):
             for col in hparams.real_columns
             if self.selected_columns is None or col in self.selected_columns
         ]
+        print(f"[INFO] {self.categorical_columns = }")
+        print(f"[INFO] {self.real_columns = }")
 
         self.target_columns = hparams.target_columns
         self.target_column_types = hparams.target_column_types
@@ -389,6 +392,7 @@ class TransformerModel(nn.Module):
         self.scheduler = self._get_scheduler(
             **self._filter_key(hparams.training_spec.scheduler, "name")
         )
+        self.scheduler_step_iter = hparams.training_spec.scheduler_step_iter
 
         self.iter_save = hparams.training_spec.iter_save
         self.continue_training = hparams.training_spec.continue_training
@@ -717,48 +721,113 @@ class TransformerModel(nn.Module):
         best_val_loss = float("inf")
         n_epochs_no_improvement = 0
 
-        for epoch in range(self.start_epoch, self.hparams.training_spec.epochs + 1):
-            if (
-                self.early_stopping_epochs is None
-                or n_epochs_no_improvement < self.early_stopping_epochs
-            ) and (
-                epoch == self.start_epoch
-                or epoch > self.start_epoch
-                and not np.isnan(total_loss)  # type: ignore # noqa: F821
-            ):
-                epoch_start_time = time.time()
+        last_epoch = None
+        best_model = None
+        try:
+            for epoch in range(self.start_epoch, self.hparams.training_spec.epochs + 1):
+                if (
+                    self.early_stopping_epochs is None
+                    or n_epochs_no_improvement < self.early_stopping_epochs
+                ) and (
+                    epoch == self.start_epoch
+                    or epoch > self.start_epoch
+                    and not np.isnan(total_loss)  # type: ignore # noqa: F821
+                ):
+                    epoch_start_time = time.time()
 
-                if train_sampler and not isinstance(train_sampler, RandomSampler):
-                    train_sampler.set_epoch(epoch)
-                self._train_epoch(train_loader, epoch)
+                    if train_sampler and not isinstance(train_sampler, RandomSampler):
+                        train_sampler.set_epoch(epoch)
+                    self._train_epoch(train_loader, epoch)
 
-                if valid_sampler and not isinstance(valid_sampler, RandomSampler):
-                    valid_sampler.set_epoch(epoch)
+                    if valid_sampler and not isinstance(valid_sampler, RandomSampler):
+                        valid_sampler.set_epoch(epoch)
 
-                total_loss, total_losses, output = self._evaluate(valid_loader)
-                elapsed = time.time() - epoch_start_time
+                    total_loss, total_losses, output = self._evaluate(valid_loader)
+                    elapsed = time.time() - epoch_start_time
 
-                self._log_epoch_results(
-                    epoch, elapsed, total_loss, total_losses, output
-                )
+                    self._log_epoch_results(
+                        epoch, elapsed, total_loss, total_losses, output
+                    )
 
-                if total_loss < best_val_loss:
-                    best_val_loss = total_loss
-                    best_model = self._copy_model()
-                    n_epochs_no_improvement = 0
+                    if total_loss < best_val_loss:
+                        best_val_loss = total_loss
+                        best_model = self._copy_model()
+                        n_epochs_no_improvement = 0
+                    else:
+                        n_epochs_no_improvement += 1
+
+                    if self.scheduler_step_iter == "epoch":
+                        self.scheduler.step()
+
+                    if epoch % self.iter_save == 0:
+                        self._save(epoch, total_loss)
+
+                    last_epoch = epoch
+        except KeyboardInterrupt:
+            self.log_file.write("\n" + "=" * 89)
+            self.log_file.write("[WARNING] Training interrupted by user (Ctrl+C).")
+
+            if self.hparams.training_spec.distributed:
+                dist.barrier()
+
+            if self.rank == 0:  # Only ask and export on the main process
+                answer = ""
+                try:
+                    answer = (
+                        input(
+                            "Do you want to export the 'best' and 'last' models? (y/n): "
+                        )
+                        .lower()
+                        .strip()
+                    )
+                except EOFError:  # Handle non-interactive environments
+                    answer = "n"
+
+                if answer == "y":
+                    self.log_file.write("[INFO] User opted to export models.")
+
+                    # Export last model (current state)
+                    if last_epoch is not None and best_model is not None:
+                        self.log_file.write(
+                            f"[INFO] Exporting 'last' model from epoch {last_epoch}..."
+                        )
+                        self._export(self, "last", last_epoch)
+
+                        # Export best model if it exists
+                        if best_model is not None:
+                            self.log_file.write(
+                                "[INFO] Exporting 'best' model (based on best val loss)..."
+                            )
+                            self._export(best_model, "best", last_epoch)
+                        else:
+                            self.log_file.write(
+                                "[WARNING] No 'best' model was saved (no validation improvement yet)."
+                            )
+
+                        self.log_file.write("[INFO] Models exported.")
+                    else:
+                        self.log_file.write(
+                            "[INFO] Could not export model as no epoch ran."
+                        )
                 else:
-                    n_epochs_no_improvement += 1
+                    self.log_file.write("[INFO] User opted *not* to export. Exiting.")
 
-                self.scheduler.step()
-                if epoch % self.iter_save == 0:
-                    self._save(epoch, total_loss)
+        if self.hparams.training_spec.distributed:
+            dist.barrier()
 
-                last_epoch = epoch
+        # Only rank 0 should perform the final export
+        if self.rank == 0:
+            if (
+                best_model is None
+            ):  # <-- MODIFICATION: Handle case where no improvement was ever made
+                self.log_file.write(
+                    "[INFO] No validation improvement during training. Saving last model as 'best'."
+                )
+                best_model = self._copy_model()
 
-        self._export(self, "last", last_epoch)  # type: ignore
-        self._export(best_model, "best", last_epoch)  # type: ignore
-        self.log_file.write("--- Training Complete ---")
-        self.log_file.close()
+            self._export(self, "last", last_epoch)  # type: ignore
+            self._export(best_model, "best", last_epoch)  # type: ignore
+            self.log_file.write("--- Training Complete ---")
 
     @beartype
     def _train_epoch(
@@ -769,7 +838,9 @@ class TransformerModel(nn.Module):
         """Trains the model for one epoch.
 
         Iterates through the training DataLoader, computes loss, performs
-        backpropagation, and updates model parameters.
+        backpropagation, and updates model parameters. The DataLoader is expected
+        to yield tuples of (data_dict, targets_dict, sequence_ids, subsequence_ids, start_positions).
+        The IDs and positions are currently unused in this training loop.
 
         Args:
             train_loader: DataLoader for the training dataset.
@@ -781,7 +852,7 @@ class TransformerModel(nn.Module):
         start_time = time.time()
         num_batches = len(train_loader)
 
-        for batch_count, (data, targets) in enumerate(train_loader):
+        for batch_count, (data, targets, _, _, _) in enumerate(train_loader):
             data = {
                 k: v.to(self.device, non_blocking=True)
                 for k, v in data.items()
@@ -820,6 +891,9 @@ class TransformerModel(nn.Module):
 
             del data, targets, output, loss, losses
 
+            if self.scheduler_step_iter == "batch":
+                self.scheduler.step()
+
     @beartype
     def _calculate_loss(
         self, output: dict[str, Tensor], targets: dict[str, Tensor]
@@ -843,7 +917,8 @@ class TransformerModel(nn.Module):
               target column.
         """
         losses = {}
-        for target_column, target_column_type in self.target_column_types.items():
+        for target_column in targets.keys():
+            target_column_type = self.target_column_types[target_column]
             if target_column_type == "categorical":
                 output[target_column] = output[target_column].reshape(
                     -1, self.n_classes[target_column]
@@ -855,7 +930,7 @@ class TransformerModel(nn.Module):
                 output[target_column], targets[target_column].T.contiguous().reshape(-1)
             )
         loss = None
-        for target_column in self.target_columns:
+        for target_column in targets.keys():
             losses[target_column] = losses[target_column] * (
                 self.loss_weights[target_column]
                 if self.loss_weights is not None
@@ -923,6 +998,9 @@ class TransformerModel(nn.Module):
         Iterates through the validation data, calculates the total loss,
         and aggregates results across all processes if in distributed mode.
         Also calculates a one-time baseline loss on the first call.
+        The DataLoader is expected to yield tuples of
+        (data_dict, targets_dict, sequence_ids, subsequence_ids, start_positions).
+        The IDs and positions are currently unused during evaluation.
 
         Args:
             valid_loader: DataLoader for the validation dataset.
@@ -940,7 +1018,7 @@ class TransformerModel(nn.Module):
         total_losses_collect = {col: [] for col in self.target_columns}
         output = {}  # for type checking
         with torch.no_grad():
-            for i, (data, targets) in enumerate(valid_loader):
+            for data, targets, _, _, _ in valid_loader:
                 # Move data to the current process's assigned GPU
                 data = {
                     k: v.to(self.device, non_blocking=True)
@@ -1000,7 +1078,7 @@ class TransformerModel(nn.Module):
             baseline_losses_local_collect = {col: [] for col in self.target_columns}
 
             # Iterate over the sharded validation loader
-            for data, targets in valid_loader:
+            for data, targets, _, _, _ in valid_loader:
                 data = {
                     k: v.to(self.device, non_blocking=True)
                     for k, v in data.items()
@@ -1011,23 +1089,32 @@ class TransformerModel(nn.Module):
                     for k, v in targets.items()
                     if k in self.target_column_types
                 }
-                # Replicate original logic of using input as pseudo-output
-                pseudo_output = {
-                    col: self._transform_val(col, data[col]) for col in targets.keys()
+
+                pseudo_output = {}
+                targets_for_baseline = {}
+                for col in self.target_columns:
+                    if col in data:
+                        pseudo_output[col] = self._transform_val(col, data[col])
+                        targets_for_baseline[col] = targets[col]
+
+                if len(pseudo_output) > 0:
+                    loss, losses = self._calculate_loss(
+                        pseudo_output, targets_for_baseline
+                    )
+                    baseline_loss_local_collect.append(loss.item())
+                    for col, loss_ in losses.items():
+                        baseline_losses_local_collect[col].append(loss_.item())
+
+                # Sum the losses for the local shard
+            if len(baseline_loss_local_collect):
+                baseline_loss_local = np.sum(baseline_loss_local_collect)
+                baseline_losses_local = {
+                    col: np.sum(loss_list)
+                    for col, loss_list in baseline_losses_local_collect.items()
                 }
-
-                loss, losses = self._calculate_loss(pseudo_output, targets)
-
-                baseline_loss_local_collect.append(loss.item())
-                for col, loss_ in losses.items():
-                    baseline_losses_local_collect[col].append(loss_.item())
-
-            # Sum the losses for the local shard
-            baseline_loss_local = np.sum(baseline_loss_local_collect)
-            baseline_losses_local = {
-                col: np.sum(loss_list)
-                for col, loss_list in baseline_losses_local_collect.items()
-            }
+            else:
+                baseline_loss_local = -1.0
+                baseline_losses_local = {col: -1.0 for col in self.target_columns}
 
             # Broadcast the baseline values from the main process to all others
             if self.hparams.training_spec.distributed:
@@ -1280,13 +1367,12 @@ class TransformerModel(nn.Module):
     def _initialize_log_file(self):
         """Initializes the log file."""
         os.makedirs(os.path.join(self.project_path, "logs"), exist_ok=True)
-        open_mode = "w" if self.start_epoch == 1 else "a"
         path = os.path.join(
             self.project_path, "logs", f"sequifier-{self.model_name}-[NUMBER].txt"
         )
         if self.rank is not None:
             path = path.replace("[NUMBER]", f"rank{self.rank}-[NUMBER]")
-        self.log_file = LogFile(path, open_mode, self.rank)
+        self.log_file = LogFile(path, self.rank)
 
     @beartype
     def _load_weights_conditional(self) -> str:
@@ -1462,7 +1548,10 @@ def load_inference_model(
                 if isinstance(module, torch.nn.Dropout):
                     module.train()
 
-        model = torch.compile(model).to(device)
+        if device.startswith("cuda"):
+            model = torch.compile(model).to(device)
+        else:
+            model.to(device)
 
     return model
 
